//    WeIRCd IRC server software - sub file
//    Copyright (C) 2007-2009 Jonas Thiem
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License v3 as published
//    by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

// ----- This is the WeIRCd API for Services  ----- //
  /* Use this to connect services bots to the
     server. You can use it to login a new fake
     user connection as a bot with services
     abilities (ths is not manually possible).
     
     Please note you need to activate usage of
     this api with weircd_infoq_ReportServicesFunctionality()
     (see function reference below).
     
     The following list of functions must be
     implemented by your module as WeIRCd will
     call them (given you activate the services
     api for your module).
     
   Function needed: void module_ServicesCOptionPreCheck_Begin();
     Will be called by WeIRCd before the config
     parser asks your module about specific
     config options being phrased correctly
     (see following function).
     
     
   Function needed: int module_ServicesCOptionPreCheck(const char* option, const char* value);
     The config parser will ask you if any option
     passed to your module is phrased correctly.
     If you return 0 it means the option is correct,
     otherwise return -1.
     Sometimes, you might not be able to fully
     evaluate the correctness of an option without
     performing intensive tests or without actually
     perform some of your module's usage. In this
     case, feel free to just return 0 and perform
     those very runtime dependent checks when the
     module is actually started with those options
     and complain then.
     You are free to always return 0, but it is
     considered way better behaviour if you
     preevaluate the options as much as possible
     at this stage to make it easier for the user
     to track config mistakes.
   
   
   Function needed: int module_ServicesCOptionPreCheck_End();
     Will be called after the config parser finished
     all its config option checking requests to your
     module.
     Together with module_ServicesCOptionPreCheck_Begin(),
     this allows you to maintain a list of all the
     provided options and when this function is
     called, you may return -1 to indicate that
     some options are missing, so the config parser
     can spit out an error respectively.
     
   Important Note: The config parser will probably not call
     this when your module reported an option to be
     invalid.
     
     
   Function needed: int module_ServicesStartup(int ocount, const char** oname, const char** ovalue);
     This function will be called by WeIRCd when it
     expects your module to startup its services
     bots. WeIRCd will pass the config options to
     it that the user specified in the
     configuration file.

     The startup of the services is completely
     independent from the module's main startup and
     termination calls. In case your module is
     not only listed as services module in the
     config this will be handled independently.

     The first value is the count of options, the
     second parameter is an array of the various
     names of the different options the user
     entered and the third is an array of the
     values the user assigned to them.

     If you consider any option or any option's
     value invalid, notify the user by issuing a
     weircd_services_StartupError() command and
     returning -1. Otherwise, do your startup and
     save your options and then return 0.
     
   Important Note: Any checks you already did on the options
     in module_ServicesCOptionPreCheck() do not
     need to be redone. That also means you will
     surely get all those options WeIRCd asked for
     in the precheck and you do not need to check
     for missing options (given you did that in the
     precheck already).
     
     
   Function needed: int module_NewOptionSet(int ocount, const char** oname, const char** ovalue)
     Due to a change of the config, some options have
     been changed. WeIRCd therefore passes you all
     the options in their new state (also if not
     all of them changed).
     
     Please take measures to enforce those new
     options before returning 0 to indicate
     success.
     
     If any of those options appear to be invalid,
     please return -1.
     
   Important Note: If you return -1, you are expected
     to perform a safe shutdown of the services.
     WeIRCd will expect your services functionality
     to be deactivated (that is, also any fake bots
     related to your services functionality removed)
     when you returned -1.
     
     
   Function needed: void module_ServicesShutdown();
     WeIRCd wants to shutdown the module's services
     functionality. Please safely close all the
     databases and remove the fake user bots that
     are related to your services functionality from
     the server before returning. */


void weircd_infoq_ReportServicesFunctionality();
  /* Use this during module_InfoQuery() to activate the services api for
     your module. */
  
int weircd_services_LoginFakeuserAsServiceBot(int handle, const char* nick, const char* userid, const char* host, const char* realname);
 /* Log in a yet not logged in fake user connection as a services bot with
    respective rights. If the nick is already taken or the handle is invalid,
    the function will return -1, otherwise 0 to indicate success.
    The function will also return -1 when the nickname, userid or host is invalid. */

void weircd_services_SetUserLoginAccountName(const char* user, const char* account);
 /* Use this to mark the target user as logged in with the account "account". Use
    NULL as account name if you want to mark the user as logged out. */

int weircd_services_NickservKill(const char* ip);
 /* Report a failed nickserv login attempt and at the same time query if
    enough attempts of that ip have been made to justify a kill to stop
    people guessing a password. */
 
void weircd_services_NickservLogin(const char* ip);
 /* Report a successful nickserv login attempt from that ip (which makes
    the server forget any unsuccessful ones having happened before). */